from __future__ import annotations

from typing import Any

from vial import timestamps
from vial.exceptions import ServerError, VialError
from vial.types import Request


class RequestContext:
    _INSTANCE: RequestContext | None = None

    def __init__(self, request: Request) -> None:
        self.request = request
        self.start_time = timestamps.epoch_millis()

    @property
    def elapsed_time(self) -> float:
        return timestamps.epoch_millis() - self.start_time

    @property
    def remaining_time(self) -> int:
        return self.request.context.get_remaining_time_in_millis()

    def __enter__(self) -> RequestContext:
        RequestContext._INSTANCE = self
        return self

    def __exit__(self, *_: Any) -> None:
        RequestContext._INSTANCE = None

    @classmethod
    def active(cls) -> RequestContext:
        if not cls._INSTANCE:
            raise ServerError(VialError.NOT_IN_REQUEST.get())
        return cls._INSTANCE


def get() -> Request:
    return RequestContext.active().request


def elapsed_time() -> float:
    return RequestContext.active().elapsed_time


def remaining_time() -> int:
    return RequestContext.active().remaining_time



from __future__ import annotations

import logging
from logging import Formatter, Handler, Logger, LogRecord, StreamHandler
from typing import Any, Type

from vial import timestamps
from vial.json import Json, NativeJson


class LoggerFactory:

    DEFAULT_FORMAT = "[%(levelname)s] %(asctime)s %(filename)s.%(funcName)s: %(message)s"

    @classmethod
    def get(cls, name: str) -> Logger:
        log = logging.getLogger(name)
        log.setLevel(logging.DEBUG)
        log.propagate = False
        log.addHandler(cls.get_handler())
        return log

    @classmethod
    def get_handler(cls) -> Handler:
        handler = StreamHandler()
        handler.setFormatter(Formatter(cls.DEFAULT_FORMAT))
        return handler


class JsonFormatter(Formatter):

    json_class: Type[Json] = NativeJson

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.json = self.json_class()

    def format(self, record: LogRecord) -> str:
        json_record = self.to_record(record)
        self.add_traceback(record, json_record)
        return self.json.dumps(json_record)

    def add_traceback(self, record: LogRecord, json_record: dict[str, str]) -> None:
        if record.exc_info:
            json_record["traceback"] = self.formatException(record.exc_info)

    def formatTime(self, record: LogRecord, datefmt: str | None = None) -> str:
        if datefmt:
            return super().formatTime(record, datefmt)
        return timestamps.pretty(timestamps.of_epoch_seconds(record.created))

    def to_record(self, record: LogRecord) -> dict[str, str]:
        return {
            "message": record.getMessage(),
            "name": record.name,
            "level": record.levelname,
            "file": record.filename,
            "module": record.module,
            "function": record.funcName,
            "time": self.formatTime(record),
        }


class JsonLoggerFactory(LoggerFactory):
    @classmethod
    def get_handler(cls) -> Handler:
        handler = StreamHandler()
        handler.setFormatter(JsonFormatter())
        return handler



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, auto
from http import HTTPStatus
from typing import Any, Iterator, MutableMapping, TypeVar

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class MultiDict(MutableMapping[K, list[V]]):  # pylint: disable=too-many-ancestors
    def __init__(self, values: dict[K, list[V]] | None = None) -> None:
        super().__init__()
        self._values = values or {}

    def __delitem__(self, key: K) -> None:
        del self._values[key]

    def __getitem__(self, key: K) -> list[V]:
        return self._values[key]

    def get_first(self, key: K) -> V:
        return self._values[key][0]

    def extend(self, key: K, value: list[V]) -> None:
        if (existing_values := self._values.get(key)) is not None:
            existing_values.extend(value)
        else:
            self._values[key] = value

    def add(self, key: K, value: V) -> None:
        if (existing_values := self._values.get(key)) is not None:
            existing_values.append(value)
        else:
            self._values[key] = [value]

    def __iter__(self) -> Iterator[K]:
        return iter(self._values)

    def __len__(self) -> int:
        return len(self._values)

    def __setitem__(self, key: K, value: list[V] | V) -> None:
        if isinstance(value, list):
            self._values[key] = value
        else:
            self._values[key] = [value]

    def __str__(self) -> str:
        return str(self._values)

    def __repr__(self) -> str:
        return repr(self._values)


@dataclass
class CognitoIdentity:
    cognito_identity_id: str
    cognito_identity_pool_id: str


@dataclass
class MobileClient:
    installation_id: str
    app_title: str
    app_version_name: str
    app_version_code: str
    app_package_name: str


@dataclass
class ClientContext:
    client: MobileClient
    custom: dict[str, Any]
    env: dict[str, Any]


@dataclass
class LambdaContext:
    function_name: str
    function_version: str
    invoked_function_arn: str
    memory_limit_in_mb: int
    aws_request_id: str
    log_group_name: str
    log_stream_name: str
    identity: CognitoIdentity | None = None
    client_context: ClientContext | None = None

    @staticmethod
    def get_remaining_time_in_millis() -> int:
        """
        This is never used at runtime, the AWS provided context is used instead,
        and this is just a placeholder for tests and type hints.
        """
        return 30_000


class HTTPMethod(Enum):
    GET = auto()
    PUT = auto()
    PATCH = auto()
    POST = auto()
    DELETE = auto()
    OPTIONS = auto()
    TRACE = auto()


@dataclass
class LambdaEvent:
    event: dict[str, Any]
    context: LambdaContext


@dataclass
class Request(LambdaEvent):
    method: HTTPMethod
    resource: str
    path: str
    headers: MultiDict[str, str]
    query_parameters: MultiDict[str, str]
    body: str | None


@dataclass
class Response:
    body: dict[str, Any] | list[Any] | str | None = None
    headers: dict[str, str] = field(default_factory=dict)
    status: HTTPStatus | int = HTTPStatus.OK



from __future__ import annotations

import dataclasses
from collections import defaultdict
from http import HTTPStatus
from typing import Callable, Type, TypeVar, cast

from vial.exceptions import HTTPError, ServerError, VialError
from vial.types import Response

E = TypeVar("E", bound=Exception)


class ErrorHandler:
    DEFAULT_STATUSES = {
        Exception: HTTPStatus.INTERNAL_SERVER_ERROR,
        ValueError: HTTPStatus.BAD_REQUEST,
        NotImplementedError: HTTPStatus.NOT_IMPLEMENTED,
    }

    def __init__(self, name: str) -> None:
        self.name = name
        self.error_handlers: dict[str, dict[Type[Exception], Callable[[Exception], Response]]] = defaultdict(dict)
        self.register_handler(Exception, self._default_handler)
        self.register_handler(HTTPError, self._http_error_handler)
        self.register_handler(ServerError, self._server_error_handler)

    def register_handler(self, error_type: Type[Exception], handler: Callable[[E], Response]) -> None:
        self.error_handlers[self.name][error_type] = cast(Callable[[Exception], Response], handler)

    def __call__(self, resource: str, error: Exception) -> Response:
        for error_type in type(error).__mro__:
            if issubclass(error_type, Exception):
                if handler := self._get_error_handler(resource, error_type):
                    return handler(error)
        return self._default_handler(error)

    def _get_error_handler(self, resource: str, error_type: Type[Exception]) -> Callable[[Exception], Response] | None:
        """
        Tries to get the best matching error handler for the specified route. An error handler registered on a
        Resource rather than the global Vial application will always take precedence over the global erro handler.
        """
        return self.error_handlers[resource].get(error_type) or self.error_handlers[self.name].get(error_type)

    def _server_error_handler(self, error: ServerError) -> Response:
        return Response(dataclasses.asdict(error.error), status=error.status)

    def _http_error_handler(self, error: HTTPError) -> Response:
        body = dataclasses.asdict(VialError.UNKNOWN_ERROR.get(str(error)))
        return Response(body, status=error.status)

    def _default_handler(self, error: Exception) -> Response:
        body = dataclasses.asdict(VialError.UNKNOWN_ERROR.get(str(error)))
        return Response(body, status=self._get_native_status_code(error))

    def _get_native_status_code(self, error: Exception) -> HTTPStatus:
        for error_type in type(error).__mro__:
            if issubclass(error_type, Exception):
                if status := self.DEFAULT_STATUSES.get(error_type):
                    return status
        return HTTPStatus.INTERNAL_SERVER_ERROR


class ErrorHandlingAPI:
    error_handler_class: Type[ErrorHandler] = ErrorHandler

    def __init__(self, name: str) -> None:
        self.name = name
        self.default_error_handler = self.error_handler_class(name)

    def error_handler(
        self, *error_types: Type[Exception]
    ) -> Callable[[Callable[[E], Response]], Callable[[E], Response]]:
        """Binds the decorated function as the error handler for all provided exception types."""

        def error_handling_wrapper(function: Callable[[E], Response]) -> Callable[[E], Response]:
            for error_type in error_types:
                self.register_error_handler(error_type, function)
            return function

        return error_handling_wrapper

    def register_error_handler(self, error_type: Type[Exception], handler: Callable[[E], Response]) -> None:
        self.default_error_handler.register_handler(error_type, handler)

    def register_error_handlers(self, other: ErrorHandlingAPI) -> None:
        self.default_error_handler.error_handlers[other.name] = other.default_error_handler.error_handlers[other.name]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from http import HTTPStatus
from typing import Any


@dataclass
class ErrorCode:
    CODE = "code"
    MESSAGE = "message"

    code: str
    message: str


class ErrorCodeBuilder:
    value: tuple[int, str]
    name: str

    def get(self, *args: Any, **kwargs: Any) -> ErrorCode:
        return ErrorCode(self.name, self.value[1].format(*args, **kwargs))


class VialError(ErrorCodeBuilder, Enum):
    ROUTE_NOT_FOUND = auto(), "No route defined for resource {}"
    METHOD_NOT_ALLOWED = auto(), "No route defined for resource {} and method {}"
    PARSER_NOT_REGISTERED = auto(), "Parser '{}' is not registered"
    PARSER_ALREADY_EXISTS = auto(), "Parser '{}' is already registered"
    NOT_IN_REQUEST = auto(), "Not currently within a request"
    INVALID_TIMESTAMP_ZONE = auto(), "Only UTC timestamps are supported, got {}"
    UNKNOWN_ERROR = auto(), "{}"


class HTTPError(Exception):
    status = HTTPStatus.INTERNAL_SERVER_ERROR


class ServerError(HTTPError):
    def __init__(self, error: ErrorCode) -> None:
        super().__init__(error.message)
        self.error = error


class BadRequestError(ServerError):
    status = HTTPStatus.BAD_REQUEST


class UnauthorizedError(ServerError):
    status = HTTPStatus.UNAUTHORIZED


class ForbiddenError(ServerError):
    status = HTTPStatus.FORBIDDEN


class NotFoundError(ServerError):
    status = HTTPStatus.NOT_FOUND


class MethodNotAllowedError(ServerError):
    status = HTTPStatus.METHOD_NOT_ALLOWED



from datetime import datetime, timezone

from vial.exceptions import ServerError, VialError

DEFAULT_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def epoch_millis() -> float:
    return now().timestamp() * 1000


def now() -> datetime:
    return datetime.now(timezone.utc)


def parse(timestamp: str) -> datetime:
    return datetime.strptime(timestamp, DEFAULT_FORMAT).replace(tzinfo=timezone.utc)


def pretty(timestamp: datetime) -> str:
    if not (timestamp.tzinfo is None or timestamp.tzinfo == timezone.utc):
        raise ServerError(VialError.INVALID_TIMESTAMP_ZONE.get(timestamp.tzinfo))
    return timestamp.strftime(DEFAULT_FORMAT)


def of_epoch_seconds(seconds: float) -> datetime:
    return datetime.fromtimestamp(seconds, timezone.utc)



import dataclasses
import json
from enum import Enum
from json.encoder import JSONEncoder
from typing import Any, Callable, Protocol, Type
from uuid import UUID


class Json(Protocol):
    @staticmethod
    def dumps(value: Any) -> str:
        pass

    @staticmethod
    def loads(value: str) -> Any:
        pass


def _enum_to_string(value: Enum) -> str:
    return value.name


def _instance_check(class_: Type[Any]) -> Callable[[Any], bool]:
    def instance_checker(value: Any) -> bool:
        return isinstance(value, class_)

    return instance_checker


class DefaultEncoder(JSONEncoder):
    ENCODERS: list[tuple[Callable[[Any], bool], Callable[[Any], Any]]] = [
        (_instance_check(set), list),
        (_instance_check(UUID), str),
        (_instance_check(Enum), _enum_to_string),
        (dataclasses.is_dataclass, dataclasses.asdict),
    ]

    def default(self, o: Any) -> Any:
        for type_matcher, encoder in self.ENCODERS:
            if type_matcher(o):
                return encoder(o)
        return super().default(o)


class NativeJson(Json):
    @staticmethod
    def dumps(value: Any) -> str:
        return json.dumps(value, cls=DefaultEncoder)

    @staticmethod
    def loads(value: str) -> Any:
        return json.loads(value)



import base64
from typing import Any, Type, cast

from vial.errors import ErrorHandlingAPI
from vial.exceptions import MethodNotAllowedError, NotFoundError, VialError
from vial.json import Json, NativeJson
from vial.loggers import LoggerFactory
from vial.middleware import CallChain, MiddlewareAPI, MiddlewareChain
from vial.parsers import ParserAPI
from vial.request import RequestContext
from vial.routes import Route, RoutingAPI
from vial.types import HTTPMethod, LambdaContext, MultiDict, Request, Response


class RouteResolver:
    """
    Rudimentary resolver which tries to find a one to one mapping between the API Gateway resource
    and a route defined within the Vial application. As an example an API Gateway may have a resource
    under the "/users/{user_id}/addresses" path, which will be matched against an exact same mapping on
    Vial, but will not be matched against "/users/something/addresses".

    This approach only works when using direct routes and won't work with Proxy+ integrations, which don't
    try to match against pre-defined resources. In such a use case, this class can be overridden with a
    custom implementation and used to replace the Vial#route_resolver_class class field. For this use case,
    the vial.gateway.RouteMatcher class can be used to match conventional URLs to route resources.
    """

    def __call__(self, resources: dict[str, dict[HTTPMethod, Route]], request: Request) -> Route:
        if not (defined_routes := resources.get(request.resource)):
            raise NotFoundError(VialError.ROUTE_NOT_FOUND.get(request.resource))

        if not (route := defined_routes.get(request.method)):
            raise MethodNotAllowedError(VialError.METHOD_NOT_ALLOWED.get(request.resource, request.method.name))

        return route


class RouteInvoker:
    def __call__(self, route: Route, request: Request) -> Response:
        """
        Invokes the route function with path parameters passed in the exact same order as they
        are defined in the route. This has the advantage of being able to rename path parameters
        in code without affecting the API Gateway integration, which may not allow renames in
        certain circumstances.

        This behaviour can be modified to bind the path parameters by name to the route function
        with kwarg style parameter passing, by overriding this class and using that as the new
        value of the Vial#route_invoker_class field.
        """
        args = self._build_args(route, request)
        result = route.function(*args.values())
        return self._to_response(result)

    @staticmethod
    def _to_response(result: Any) -> Response:
        if isinstance(result, Response):
            return result
        if result is None or not isinstance(result, tuple):
            return Response(result)
        return Response(*result)

    @staticmethod
    def _build_args(route: Route, request: Request) -> dict[str, Any]:
        if not route.variables:
            return {}
        path_params: dict[str, str] = request.event["pathParameters"]
        args: dict[str, Any] = {}
        for name, parser in route.variables.items():
            args[name] = parser(path_params[name])
        return args


class Resource(RoutingAPI, ParserAPI, MiddlewareAPI, ErrorHandlingAPI):
    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.name = name


class Vial(RoutingAPI, ParserAPI, MiddlewareAPI, ErrorHandlingAPI):

    route_resolver_class = RouteResolver

    route_invoker_class = RouteInvoker

    logger_factory_class = LoggerFactory

    json_class: Type[Json] = NativeJson

    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.name = name
        self.route_resolver = self.route_resolver_class()
        self.invoker = self.route_invoker_class()
        self.json = self.json_class()
        self.logger = self.logger_factory_class.get(name)

    def register_resource(self, app: Resource) -> None:
        self.register_parsers(app)
        self.register_routes(app)
        self.register_middlewares(app)
        self.register_error_handlers(app)

    def __call__(self, event: dict[str, Any], context: LambdaContext) -> dict[str, Any]:
        request = self._build_request(event, context)
        with RequestContext(request):
            response = self._handle_request(request)
            return self._to_lambda_response(response)

    def _handle_request(self, request: Request) -> Response:
        route_resource = self.name  # If a route can't be found, default to the global application
        try:
            route = self.route_resolver(self.routes, request)
            route_resource = route.resource
            return self._build_invocation_chain(route)(request)
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception("Encountered uncaught exception")
            return self.default_error_handler(route_resource, e)

    def _build_invocation_chain(self, route: Route) -> CallChain:
        def route_invocation(event: Request) -> Response:
            return self.invoker(route, event)

        if not (all_middleware := self.registered_middleware[self.name] + self.registered_middleware[route.resource]):
            return route_invocation

        handler = MiddlewareChain(all_middleware[-1], route_invocation)
        for middleware in reversed(all_middleware[0:-1]):
            handler = MiddlewareChain(middleware, handler)
        return handler

    def _to_lambda_response(self, response: Response) -> dict[str, Any]:
        if not isinstance(response.body, str):
            body = self.json.dumps(response.body) if response.body is not None else None
        else:
            body = response.body
        return {"headers": response.headers, "statusCode": response.status, "body": body}

    def _build_request(self, event: dict[str, Any], context: LambdaContext) -> Request:
        return Request(
            event,
            context,
            HTTPMethod[event["httpMethod"]],
            event["resource"],
            event["path"],
            MultiDict(event["multiValueHeaders"]),
            MultiDict(event["multiValueQueryStringParameters"]),
            self._get_body(event),
        )

    @staticmethod
    def _get_body(event: dict[str, Any]) -> str:
        body = event["body"]
        if event.get("isBase64Encoded") and body:
            return base64.b64decode(body).decode("utf-8")
        return cast(str, body)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Type
from urllib import parse

from vial.app import Vial
from vial.exceptions import NotFoundError, VialError
from vial.json import Json, NativeJson
from vial.types import HTTPMethod, LambdaContext, Response


@dataclass
class Match:
    route: str
    path_params: dict[str, str]
    query_params: dict[str, list[str]]


class RouteMatcher:
    def __init__(self, routes: list[str]) -> None:
        self.routes = sorted(routes)

    def match(self, url: str) -> Match:
        parsed_url = parse.urlparse(url)
        query_params = parse.parse_qs(parsed_url.query, keep_blank_values=True)
        path = self._remove_trailing_slash(parsed_url.path)
        return self._find_match(url, path.split("/"), query_params)

    def _find_match(self, url: str, parts: list[str], query_params: dict[str, list[str]]) -> Match:
        for route_url in self.routes:
            url_parts = route_url.split("/")
            if len(parts) != len(url_parts):
                continue

            if (path_params := self._match_path(parts, url_parts)) is not None:
                return Match(route_url, path_params, query_params)
        raise NotFoundError(VialError.ROUTE_NOT_FOUND.get(url))

    @staticmethod
    def _match_path(parts: list[str], url_parts: list[str]) -> dict[str, str] | None:
        path_params: dict[str, str] = {}
        for left, right in zip(parts, url_parts):
            if right.startswith("{") and right.endswith("}"):
                path_params[right[1:-1]] = left
                continue

            if left != right:
                return None
        return path_params

    @staticmethod
    def _remove_trailing_slash(path: str) -> str:
        if path != "/" and path.endswith("/"):
            return path[:-1]
        return path


class Gateway:

    json_class: Type[Json] = NativeJson

    def __init__(self, app: Vial) -> None:
        self.app = app
        self.json = self.json_class()
        self.matcher = RouteMatcher(list(app.routes))

    def get(self, path: str, headers: dict[str, str | list[str]] | None = None) -> Response:
        return self.request(HTTPMethod.GET, path, headers=headers)

    def post(self, path: str, body: str | None = None, headers: dict[str, str | list[str]] | None = None) -> Response:
        return self.request(HTTPMethod.POST, path, body, headers)

    def put(self, path: str, body: str | None = None, headers: dict[str, str | list[str]] | None = None) -> Response:
        return self.request(HTTPMethod.PUT, path, body, headers)

    def patch(self, path: str, body: str | None = None, headers: dict[str, str | list[str]] | None = None) -> Response:
        return self.request(HTTPMethod.PATCH, path, body, headers)

    def delete(self, path: str, headers: dict[str, str | list[str]] | None = None) -> Response:
        return self.request(HTTPMethod.DELETE, path, headers=headers)

    def request(
        self, method: HTTPMethod, path: str, body: str | None = None, headers: dict[str, str | list[str]] | None = None
    ) -> Response:
        request = self.build_request(method, path, body, headers)
        response = self.app(request, self.get_context())
        return self.build_response(response)

    def build_response(self, response: dict[str, Any]) -> Response:
        body: str | None = response["body"]
        return Response(self.json.loads(body) if body else None, response["headers"], response["statusCode"])

    def build_request(
        self,
        method: HTTPMethod,
        path: str,
        body: str | None = None,
        headers: dict[str, str | list[str]] | None = None,
    ) -> dict[str, Any]:
        match = self.matcher.match(path)
        return {
            "httpMethod": method.name,
            "resource": match.route,
            "path": path,
            "multiValueHeaders": self._build_headers(headers or {}),
            "multiValueQueryStringParameters": match.query_params,
            "pathParameters": match.path_params,
            "body": body,
        }

    @staticmethod
    def _build_headers(headers: dict[str, str | list[str]]) -> dict[str, list[str]]:
        multi_headers: dict[str, list[str]] = defaultdict(list)
        for name, value in headers.items():
            if isinstance(value, str):
                multi_headers[name].append(value)
            else:
                multi_headers[name].extend(value)
        return multi_headers

    @staticmethod
    def get_context() -> LambdaContext:
        return LambdaContext("vial-test", "1", "arn:vial-test", 128, "1", "vial-test-log", "vial-test-log")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vial']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pyvial',
    'version': '1.2.0',
    'description': 'A micro web framework for AWS Lambda.',
    'long_description': '# Vial\n<p>\n    <a href="https://pypi.org/project/pyvial/">\n        <img src="https://badgen.net/pypi/v/pyvial" alt="Latest Version" style="max-width:100%;">\n    </a>\n    <a href="https://github.com/michaeldimchuk/pyvial/actions/workflows/tests.yaml">\n        <img src="https://github.com/michaeldimchuk/pyvial/actions/workflows/tests.yaml/badge.svg" alt="Test Status" style="max-width:100%;">\n    </a>\n    <a href="https://github.com/pre-commit/pre-commit">\n        <img src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white" alt="Pre-Commit Enabled" style="max-width:100%;">\n    </a>\n</p>\n\nVial is an unopinionated micro web framework for AWS Lambda. It\'s main purpose is to provide an easy to use\ninterface while also making the core functionality of the framework as modular as possible.\n\n## Installation\nTo add vial to your project, run the following command:\n```\npoetry add pyvial\n```\n\n## Usage\n### Entry Point\nThe main entry point of the application is always the `Vial#__call__` function. When deploying to AWS Lambda,\nthe Lambda handler should point to the `Vial` object in whichever file it\'s defined in. As an example:\n```\nfrom vial.app import Vial\n\napp = Vial(__name__)\n```\nIf this code snippet is defined in an `app.py` file, the handler would be `app.app`.\n\n### Basic API\n```\nfrom vial.app import Vial\n\napp = Vial(__name__)\n\n\n@app.get("/health")\ndef health() -> dict[str, str]:\n    return {"status": "OK"}\n```\nA test case with this example is available in [tests/samples/test_with_app.py](tests/samples/test_with_app.py).\n\nThe project can be deployed with the [serverless framework](https://www.serverless.com/framework/docs/getting-started).\nTo get started with serverless for the first time, run the following commands in your project directory:\n```\nyarn add --dev serverless serverless-python-requirements\n```\n\nBelow is a basic `serverless.yml` file to configure and deploy the project. It assumes that the only file to deploy\nis app.py, and the Vial object is defined there as a variable named `app`.\n```\nservice: store\n\ncustom:\n  pythonRequirements:\n    usePoetry: true\n    slim: true\n\nprovider:\n  name: aws\n  runtime: python3.9\n  memorySize: 128\n  region: us-west-2\n\nfunctions:\n  api:\n    handler: app.app\n    events:\n      - http: get /health\n\nplugins:\n  - serverless-python-requirements\n\npackage:\n  patterns:\n    - "app.py"\n    - "!.*/**"\n    - "!package.json"\n    - "!yarn.json"\n    - "!node_modules"\n    - "poetry.lock"\n    - "pyproject.toml"\n```\nYou can now deploy the project with `yarn run serverless deploy`.\n\nAn example of this application is available at [samples/store/serverless.yml](samples/store/serverless.yml).\n\n### Current Request\nThe current request is tracked within a contextual object that wraps the lambda request, and can be accessed\nthrough the `vial.request` module. The `request.get()` function is only available during a lambda request,\nand will raise an error if called outside of one. It can be accessed like so:\n```\nfrom vial import request\nfrom vial.app import Vial\nfrom vial.types import Request\n\napp = Vial(__name__)\n\n\n@app.get("/hello-world")\ndef hello_world() -> dict[str, list[str]]:\n    if not (query_params := request.get().query_parameters):\n        raise ValueError("Must provide at least one query parameter")\n    return dict(query_params)\n```\nA test case with this example is available in [tests/samples/test_with_current_request.py](tests/samples/test_with_current_request.py).\n\n### Path Parameters\nYou can define path parameters like this:\n```\nfrom dataclasses import dataclass\n\nfrom vial.app import Vial\n\napp = Vial(__name__)\n\n\n@dataclass\nclass User:\n    user_id: str\n\n\n@app.get("/users/{user_id}")\ndef get_user(user_id: str) -> User:\n    return User(user_id)\n```\nA test case with this example is available in [tests/samples/test_with_path_parameters.py](tests/samples/test_with_path_parameters.py).\n\nVial supports some path parameter parsing as part of the invocation process. For example when using a UUID\nas a path parameter, Vial can convert it from a string to a UUID automatically:\n```\nfrom dataclasses import dataclass\nfrom uuid import UUID, uuid4\n\nfrom vial.app import Vial\n\napp = Vial(__name__)\n\n\n@dataclass\nclass User:\n    user_id: UUID\n\n\n@app.get("/users/{user_id:uuid}")\ndef get_user(user_id: UUID) -> User:\n    if not isinstance(user_id, UUID):\n        raise AssertionError("Invalid input")\n    return User(user_id)\n```\nA test case with this example is available in [tests/samples/test_with_parser.py](tests/samples/test_with_parser.py).\n\nThe following parsers are supported by default:\n\n| Parser        | Type              |\n| ------------- | ----------------- |\n| `str`         | `str`             |\n| `bool`        | `bool`            |\n| `int`         | `int`             |\n| `float`       | `float`           |\n| `decimal`     | `decimal.Decimal` |\n| `uuid`        | `uuid.UUID`       |\n\nYou can register your own parser that consumes a string variable and converts it to any other type.\nAs parsers are bound directly to the registered route function, they have to be defined before the route\nfunction that uses one is registered.\n```\n@app.parser("list")\ndef list_parser(value: str) -> list[str]:\n    return [value]\n\n\n@app.get("/users/{user_ids:list}")\ndef get_users(user_ids: list[str]) -> list[User]:\n    if not isinstance(user_ids, list) or len(user_ids) != 1:\n        raise AssertionError("Invalid input")\n    return list(map(User, user_ids))\n```\nA test case with this example is available in [tests/samples/test_with_custom_parser.py](tests/samples/test_with_custom_parser.py).\n\n## Resources\nAs your application grows, you may want to split certain functionality amongst resources and files, similar to\nblueprints of other popular frameworks like Flask.\n\nYou can define a resource like this:\n```\nfrom dataclasses import dataclass\n\nfrom vial.app import Resource, Vial\n\nstores_app = Resource(__name__)\n\n\n@dataclass\nclass Store:\n    store_id: str\n\n\n@stores_app.get("/stores/{store_id}")\ndef get_store(store_id: str) -> Store:\n    return Store(store_id)\n\n\napp = Vial(__name__)\n\napp.register_resource(stores_app)\n```\nA test case with this example is available in [tests/samples/test_with_resources.py](tests/samples/test_with_resources.py).\n\n## Middleware\nYou can register middleware functions to be executed before / after route invocations. All middleware is scoped to\nwhere it\'s registered. A middleware function registered with the `Vial` instance is scoped to all routes within\nthe application, but a function registered with a `Resource` instance will only be invoked for routes defined in\nthat specific resource.\n\nThe route invocation will be the last callable in the call chain, so any middleware in the chain before can\nexit the request and prevent the route invocation or even any other middleware from being called, exiting instead.\nThis may be helpful in cases where an application needs to fail early if the request isn\'t valid, like if a\nrequired header is missing.\n\nBelow is an example of registering a middleware to log route invocation:\n```\nfrom __future__ import annotations\n\nfrom vial import request\nfrom vial.app import Vial\nfrom vial.middleware import CallChain\nfrom vial.types import Request, Response\n\napp = Vial(__name__)\n\n\n@app.middleware\ndef log_events(event: Request, chain: CallChain) -> Response:\n    app.logger.info("Began execution of %s", event.context)\n    event.headers["custom-injected-header"] = "hello there"\n    try:\n        return chain(event)\n    finally:\n        app.logger.info("Completed execution of %s", event.context)\n\n\n@app.get("/hello-world")\ndef hello_world() -> dict[str, str | list[str]]:\n    return {"hello": "world", **request.get().headers}\n```\nA test case with this example is available in [tests/samples/test_with_middleware.py](tests/samples/test_with_middleware.py).\n\n\n## Error Handling\nWhen errors are raised by the application, the default error handler will iterate the class inheritance hierarchy of the\nexception that was raised, trying to find the most fine grained error handler possible. Default error handlers for common\nexception types like `Exception` or `ValueError` are provided, but can be overridden. Below is a sample on how to register\ncustom error handlers or override existing ones:\n```\nfrom http import HTTPStatus\n\nfrom vial.app import Vial\nfrom vial.gateway import Gateway\nfrom vial.types import Response\n\napp = Vial(__name__)\n\n\nclass CustomError(Exception):\n    pass\n\n\nclass ConfusedError(CustomError):\n    pass\n\n\n@app.error_handler(CustomError)\ndef custom_error_handler(error: CustomError) -> Response:\n    return Response({"custom_message": str(error)}, status=HTTPStatus.IM_A_TEAPOT)\n\n\n@app.error_handler(ConfusedError)\ndef confused_error_handler(error: ConfusedError) -> Response:\n    return Response({"custom_message": str(error)}, status=HTTPStatus.BAD_GATEWAY)\n\n\n@app.get("/teapot")\ndef teapot() -> None:\n    raise CustomError("I really am a teapot")\n\n\n@app.get("/confused-teapot")\ndef confused_teapot() -> None:\n    raise ConfusedError("I\'m a really confused teapot")\n```\nA test case with this example is available in [tests/samples/test_with_error_handling.py](tests/samples/test_with_error_handling.py).\n\nError handlers are bound to the resource they were registered in, whether that\'s the global `Vial` application or\na specific `Resource` instance. When an error occurs in a route, the "owner" application / resource is taken into consideration when\nchoosing the error handler to use.\n\nAn error handler registered in a `Resource` will always have precedence over a global error handler registered in the `Vial`\napplication. This allows for resources to either override global error handling mechanisms or add customization for more\nfine grained exception types.\n\nNote that the most fine grained error handler is always chosen, no matter where it comes from. That means that in a scenario\nlike this:\n```\nclass First(Exception):\n    pass\n\nclass Second(First):\n    pass\n\nclass Third(Second):\n    pass\n```\nIf the `Vial` application registers error handlers for `First` and `Third` while the `Resource` registers an override for\n`Second`, then when an exception of type `Third` is thrown, the global error handler will be used because it has a closer\nmatch to the exception, even if its parent is overridden in the `Resource`.\n\nBelow is an example of a `Resource` specific error handler:\n```\nfrom http import HTTPStatus\n\nfrom vial.app import Resource, Vial\nfrom vial.types import Response\n\napp = Vial(__name__)\n\nconfused_app = Resource(f"confused_{__name__}")\n\n\nclass CustomError(Exception):\n    pass\n\n\nclass ConfusedError(CustomError):\n    pass\n\n\n@app.error_handler(CustomError)\ndef custom_error_handler(error: CustomError) -> Response:\n    return Response({"custom_message": str(error)}, status=HTTPStatus.IM_A_TEAPOT)\n\n\n@confused_app.error_handler(ConfusedError)\ndef confused_error_handler(error: ConfusedError) -> Response:\n    return Response({"custom_message": str(error)}, status=HTTPStatus.BAD_GATEWAY)\n\n\n@app.get("/teapot")\ndef teapot() -> None:\n    raise CustomError("I really am a teapot")\n\n\n@confused_app.get("/confused-teapot")\ndef confused_teapot() -> None:\n    raise ConfusedError("I\'m a really confused teapot")\n\n\napp.register_resource(confused_app)\n```\nA test case with this example is available in [tests/samples/test_with_resource_error_handling.py](tests/samples/test_with_resource_error_handling.py).\n\n\n## Json Encoding\nYou can customize how Vial serializes / deserializes JSON objects by passing a custom encoder. The below\nexample shows how to substitute the native JSON module with another library like `simplejson`:\n```\nfrom decimal import Decimal\nfrom typing import Any\n\nimport simplejson\n\nfrom vial.app import Vial\nfrom vial.json import Json\n\n\nclass SimpleJson(Json):\n    @staticmethod\n    def dumps(value: Any) -> str:\n        return simplejson.dumps(value)\n\n    @staticmethod\n    def loads(value: str) -> Any:\n        return simplejson.loads(value)\n\n\nclass SimpleJsonVial(Vial):\n    json_class = SimpleJson\n\n\napp = SimpleJsonVial(__name__)\n\n\n@app.get("/prices")\ndef get_prices() -> dict[str, Decimal]:\n    # Decimal is not supported natively by the json module, but is by simplejson.\n    return {"bread": Decimal("42.24"), "cheese": Decimal("129.34")}\n```\nA test case with this example is available in [tests/samples/test_with_json_encoding.py](tests/samples/test_with_json_encoding.py).\n\n## Testing\nThe `vial.gateway.Gateway` class provides functionality to interact with the Vial application locally,\nwithout deploying to AWS Lambda. It can be constructed using the original `Vial` application instance,\nexposing the application endpoints with basic URL path matching.\n\nHere is an example test case using `pytest`:\n```\nfrom http import HTTPStatus\n\nimport pytest\n\nfrom vial import request\nfrom vial.app import Vial\nfrom vial.exceptions import BadRequestError\nfrom vial.gateway import Gateway\n\napp = Vial(__name__)\n\n\n@app.get("/health")\ndef health() -> dict[str, str]:\n    return {"status": "OK"}\n\n\n@app.post("/stores/{store_id}")\ndef create_store(store_id: str) -> dict[str, str]:\n    if not (body := request.get().body):\n        raise BadRequestError("Bad request")\n    return {"store_id": store_id, **app.json.loads(body)}\n\n\n@pytest.fixture(name="gateway")\ndef gateway_fixture() -> Gateway:\n    return Gateway(app)\n\n\ndef test_health(gateway: Gateway) -> None:\n    response = gateway.get("/health")\n    assert response.status == HTTPStatus.OK\n    assert response.body == {"status": "OK"}\n\n\ndef test_create_store(gateway: Gateway) -> None:\n    body = app.json.dumps({"store_name": "My cool store"})\n    response = gateway.post("/stores/my-cool-store", body)\n    assert response.status == HTTPStatus.OK\n    assert response.body == {"store_id": "my-cool-store", "store_name": "My cool store"}\n```\nThis code is also available in [tests/samples/test_with_gateway.py](tests/samples/test_with_gateway.py).\n',
    'author': 'Michael Dimchuk',
    'author_email': 'michaeldimchuk@gmail.com',
    'maintainer': 'Michael Dimchuk',
    'maintainer_email': 'michaeldimchuk@gmail.com',
    'url': 'https://github.com/michaeldimchuk/pyvial',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from decimal import Decimal
from typing import Any, Callable
from uuid import UUID

from vial.exceptions import ServerError, VialError
from vial.types import T

Parser = Callable[[str], Any]


class KeywordParser:
    """
    A path parameter parser that converts string parameters into more specific types
    prior to being injected into route handler functions.
    """

    DEFAULT_PARSERS: dict[str, Parser] = {
        "str": str,
        "bool": bool,
        "int": int,
        "float": float,
        "decimal": Decimal,
        "uuid": UUID,
    }

    def __init__(self) -> None:
        self.parsers = dict(self.DEFAULT_PARSERS)

    def get(self, name: str) -> Parser:
        if not (parser := self.parsers.get(name)):
            raise ServerError(VialError.PARSER_NOT_REGISTERED.get(name))
        return parser

    def register(self, name: str, parser: Parser) -> None:
        if name in self.parsers:
            raise ServerError(VialError.PARSER_ALREADY_EXISTS.get(name))
        self.parsers[name] = parser


class ParserAPI:
    parser_class = KeywordParser

    def __init__(self, name: str) -> None:
        super().__init__(name)  # type: ignore[call-arg] # https://github.com/python/mypy/issues/4335
        self.name = name
        self.param_parser = self.parser_class()

    def parser(self, name: str) -> Callable[[Callable[[str], T]], Callable[[str], T]]:
        def registration_function(function: Callable[[str], T]) -> Callable[[str], T]:
            self.register_parser(name, function)
            return function

        return registration_function

    def register_parser(self, name: str, parser: Callable[[str], T]) -> None:
        self.param_parser.register(name, parser)

    def register_parsers(self, other: ParserAPI) -> None:
        self.param_parser.parsers.update(other.param_parser.parsers)
